# Copyright 2018-2021
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Multimodal data features types and query mechanisms."""

from . import (
    connectivity,
    tabular,
    image,
    dataset,
)


from .feature import Feature
get = Feature.match


TYPES = Feature._get_subclasses()


def __dir__():
    return list(Feature.CATEGORIZED.keys())


def __getattr__(attr: str):
    if attr in Feature.CATEGORIZED:
        return Feature.CATEGORIZED[attr]
    else:
        hint = ""
        if isinstance(attr, str):
            import difflib
            closest = difflib.get_close_matches(attr, list(__dir__()), n=3)
            if len(closest) > 0:
                hint = f"Did you mean {' or '.join(closest)}?"
        raise AttributeError(f"No such attribute: {__name__}.{attr} " + hint)


def warm_cache():
    """Preload preconfigured multimodal data features."""
    for ftype in TYPES.values():
        _ = ftype.get_instances()



# Copyright 2018-2021
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Non-preconfigured data features hosted at EBRAINS."""

from zipfile import ZipFile
from .. import anchor as _anchor
from .. import feature

from ...retrieval import datasets


DOI_TMPL = """
doi
---
{doi}
"""


class EbrainsDataFeature(feature.Feature, category="other"):
    def __init__(
        self,
        dataset_version_id: str,
        anchor: _anchor.AnatomicalAnchor
    ):
        feature.Feature.__init__(
            self,
            modality=None,  # lazy implementation below
            description=None,  # lazy implementation below
            anchor=anchor,
            datasets=[datasets.EbrainsV3DatasetVersion(id=dataset_version_id)],
        )
        self.version = None
        self._next = None
        self._prev = None

    @property
    def id(self):
        return self._dataset.id

    @property
    def _dataset(self) -> datasets.EbrainsV3DatasetVersion:
        assert len(self.datasets) == 1
        return self.datasets[0]

    @property
    def description(self) -> str:
        return self._dataset.description

    @property
    def name(self):
        if self._dataset.name.startswith(" "):
            return f"Ebrains Dataset: {self._dataset.is_version_of[0].name}"
        else:
            return f"Ebrains Dataset: {self._dataset.name}"

    @property
    def version_identifier(self):
        return self._dataset.version_identifier

    @property
    def version_history(self):
        return self._dataset.version_changes

    @property
    def url(self):
        return self._dataset.ebrains_page

    def __hash__(self):
        return hash(self._dataset)

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, EbrainsDataFeature):
            return False
        return self._dataset == o._dataset

    def _export(self, fh: ZipFile):
        super()._export(fh)
        fh.writestr("doi.md", DOI_TMPL.format(doi=self.url))



import math
import struct
from functools import wraps
from typing import Callable

cipher = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz_-'
separator = '.'
neg = '~'
def encode_number(n, float_flag=False):
    if float_flag:
        b=struct.pack('f', n)
        new_n=struct.unpack('i',b)
        return encode_int(new_n[0])
    else:
        return encode_int(n)

def encode_int(n):
    if not isinstance(n, int):
        raise ValueError('Cannot encode int')

    residual=None
    result=''
    if n < 0:
        result += neg
        residual = n * -1
    else:
        residual = n
    
    while True:
        result = cipher[residual % 64] + result
        residual = math.floor(residual / 64)

        if residual == 0:
            break
    return result

def decode_int(n):
    neg_flag = False
    if n[-1] == neg:
        neg_flag = True
        n = n[:-1]

    result = 0
    for char in n:
        val = cipher.index(char)
        result = result * 64 + val

    if neg_flag:
        result = result * -1
    return result

def decode_number(n, float_flag=False):
    if float_flag:
        raise NotImplementedError
    return decode_int(n)


def post_process(post_process: Callable):
    def outer(fn):
        @wraps(fn)
        def inner(*args, **kwargs):
            val = fn(*args, **kwargs)
            return post_process(val, *args, **kwargs)
        return inner
    return outer



# Copyright 2018-2021
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Metadata connection to EBRAINS datasets."""

from .requests import MultiSourcedRequest, GitlabProxy, GitlabProxyEnum

import re
from typing import Union, List
from abc import ABC, abstractproperty
from hashlib import md5

try:
    from typing import TypedDict
except ImportError:
    # support python 3.7
    from typing_extensions import TypedDict


class EbrainsDatasetUrl(TypedDict):
    url: str


EbrainsDatasetPerson = TypedDict("EbrainsDatasetPerson", {
    "@id": str,
    "schema.org/shortName": str,
    "identifier": str,
    "shortName": str,
    "name": str,
})

EbrainsDatasetEmbargoStatus = TypedDict("EbrainsDatasetEmbargoStatus", {
    "@id": str,
    "name": str,
    "identifier": List[str]
})


class EbrainsBaseDataset(ABC):
    @abstractproperty
    def id(self) -> str:
        raise NotImplementedError

    @abstractproperty
    def name(self) -> str:
        raise NotImplementedError

    @abstractproperty
    def LICENSE(self) -> List[str]:
        raise NotImplementedError

    @abstractproperty
    def urls(self) -> List[EbrainsDatasetUrl]:
        raise NotImplementedError

    @abstractproperty
    def description(self) -> str:
        raise NotImplementedError

    @abstractproperty
    def contributors(self) -> List[EbrainsDatasetPerson]:
        raise NotImplementedError

    @abstractproperty
    def ebrains_page(self) -> str:
        raise NotImplementedError

    @abstractproperty
    def custodians(self) -> List[EbrainsDatasetPerson]:
        raise NotImplementedError

    def __hash__(self):
        return hash(self.id)

    def __eq__(self, o: object) -> bool:
        return hasattr(o, "id") and self.id == o.id

    def match(self, spec: Union[str, 'EbrainsBaseDataset']) -> bool:
        """
        Checks if the given specification describes this dataset.

        Parameters
        ----------
        spec (str, EbrainsBaseDataset)
            specification to be matched.
        Returns
        -------
        bool
        """
        if spec is self:
            return True
        if isinstance(spec, str):
            return self.id == spec
        raise RuntimeError(
            f"Cannot match {spec.__class__}, must be either str or EbrainsBaseDataset"
        )


class EbrainsDataset(EbrainsBaseDataset):
    """Ebrains dataset v1 connection"""

    def __init__(self, id, name=None, embargo_status: List[EbrainsDatasetEmbargoStatus] = None, *, cached_data=None):
        super().__init__()

        self._id = id
        self._name = name
        self._cached_data = cached_data
        self.embargo_status = embargo_status

        if id is None:
            raise TypeError("Dataset id is required")

        match = re.search(r"([a-f0-9-]+)$", id)
        if not match:
            raise ValueError(
                f"{self.__class__.__name__} initialized with invalid id: {self.id}"
            )

    @property
    def id(self) -> str:
        return self._id

    @property
    def detail(self):
        if not self._cached_data:
            match = re.search(r"([a-f0-9-]+)$", self.id)
            instance_id = match.group(1)
            self._cached_data = MultiSourcedRequest(
                requests=[
                    GitlabProxy(
                        GitlabProxyEnum.DATASET_V1,
                        instance_id=instance_id,
                    ),
                ]
            ).data
        return self._cached_data

    @property
    def name(self) -> str:
        if self._name is None:
            self._name = self.detail.get("name")
        return self._name

    @property
    def urls(self) -> List[EbrainsDatasetUrl]:
        return [
            {
                "url": f if f.startswith("http") else f"https://doi.org/{f}",
            }
            for f in self.detail.get("kgReference", [])
        ]

    @property
    def description(self) -> str:
        return self.detail.get("description")

    @property
    def contributors(self) -> List[EbrainsDatasetPerson]:
        return self.detail.get("contributors")

    @property
    def ebrains_page(self):
        return f"https://search.kg.ebrains.eu/instances/{self.id}"

    @property
    def custodians(self) -> EbrainsDatasetPerson:
        return self.detail.get("custodians")

    @property
    def LICENSE(self) -> str:
        return self.detail.get("license", "No license information is found.")


class EbrainsV3DatasetVersion(EbrainsBaseDataset):
    @staticmethod
    def parse_person(d: dict) -> EbrainsDatasetPerson:
        assert "https://openminds.ebrains.eu/core/Person" in d.get("type"), "Cannot convert a non person to a person dict!"
        _id = d.get("id")
        name = f"{d.get('givenName')} {d.get('familyName')}"
        return {
            '@id': _id,
            'schema.org/shortName': name,
            'identifier': _id,
            'shortName': name,
            'name': name
        }

    def __init__(self, id, *, cached_data=None) -> None:
        super().__init__()

        self._id = id
        self._cached_data = cached_data

    @property
    def detail(self):
        if not self._cached_data:
            match = re.search(r"([a-f0-9-]+)$", self._id)
            instance_id = match.group(1)
            self._cached_data = MultiSourcedRequest(
                requests=[
                    GitlabProxy(
                        GitlabProxyEnum.DATASETVERSION_V3,
                        instance_id=instance_id,
                    ),
                ]
            ).data
        return self._cached_data

    @property
    def id(self) -> str:
        return self._id

    @property
    def name(self) -> str:
        fullname = self.detail.get("fullName")
        if not fullname:
            for dataset in self.is_version_of:
                if fullname:
                    break
                fullname = dataset.name
        version_id = self.detail.get("versionIdentifier")
        return f"{fullname} ({version_id})"

    @property
    def is_version_of(self):
        if not hasattr(self, "_is_version_of"):
            self._is_version_of = [EbrainsV3Dataset(id=id.get("id")) for id in self.detail.get("isVersionOf", [])]
        return self._is_version_of

    @property
    def urls(self) -> List[EbrainsDatasetUrl]:
        return [{
            "url": doi.get("identifier", None)
        } for doi in self.detail.get("doi", [])]

    @property
    def description(self) -> str:
        description = self.detail.get("description")
        for ds in self.is_version_of:
            if description:
                break
            description = ds.description
        return description or ""

    @property
    def contributors(self) -> List[EbrainsDatasetPerson]:
        return [EbrainsV3DatasetVersion.parse_person(d) for d in self.detail.get("author", [])]

    @property
    def ebrains_page(self) -> str:
        if len(self.urls) > 0:
            return self.urls[0].get("url")
        return None

    @property
    def custodians(self) -> EbrainsDatasetPerson:
        return [EbrainsV3DatasetVersion.parse_person(d) for d in self.detail.get("custodian", [])]

    @property
    def version_changes(self):
        return self.detail.get("versionInnovation", "")

    @property
    def version_identifier(self):
        return self.detail.get("versionIdentifier", "")

    @property
    def LICENSE(self) -> str:
        return self.detail.get("license", "No license information is found.")


class EbrainsV3Dataset(EbrainsBaseDataset):
    def __init__(self, id, *, cached_data=None) -> None:
        super().__init__()

        self._id = id
        self._cached_data = cached_data
        self._contributers = None

    @property
    def id(self) -> str:
        return self._id

    @property
    def name(self) -> str:
        return self.detail.get("fullName")

    @property
    def urls(self) -> List[EbrainsDatasetUrl]:
        return [{
            "url": doi.get("identifier", None)
        } for doi in self.detail.get("doi", [])]

    @property
    def detail(self):
        if not self._cached_data:
            match = re.search(r"([a-f0-9-]+)$", self._id)
            instance_id = match.group(1)
            self._cached_data = MultiSourcedRequest(
                requests=[
                    GitlabProxy(
                        GitlabProxyEnum.DATASET_V3,
                        instance_id=instance_id,
                    ),
                ]
            ).data
        return self._cached_data

    @property
    def description(self) -> str:
        return self.detail.get("description", "")

    @property
    def contributors(self):
        if self._contributers is None:
            contributers = {}
            for version_id in self.version_ids:
                contributers.update(
                    {c['@id']: c for c in EbrainsV3DatasetVersion(version_id).contributors}
                )
            self._contributers = list(contributers.values())
        return self._contributers

    @property
    def ebrains_page(self) -> str:
        if len(self.urls) > 0:
            return self.urls[0].get("url")
        return None

    @property
    def custodians(self) -> EbrainsDatasetPerson:
        return [EbrainsV3DatasetVersion.parse_person(d) for d in self.detail.get("custodian", [])]

    @property
    def version_ids(self) -> List['str']:
        return [version.get("id") for version in self.detail.get("versions", [])]

    @property
    def LICENSE(self) -> str:
        return self.detail.get("license", "No license information is found.")


class GenericDataset():

    def __init__(
        self,
        name: str = None,
        contributors: List[str] = None,
        url: str = None,
        description: str = None,
        license: str = "No license information is found."
    ):
        self._name = name
        self._contributors = contributors
        self._url = url
        self._description = description
        self._license = license

    @property
    def contributors(self):
        return [{"name": cont} for cont in self._contributors]

    @property
    def id(self) -> str:
        return md5(self.name.encode('utf-8')).hexdigest()

    @property
    def name(self) -> str:
        return self._name

    @property
    def LICENSE(self) -> str:
        return self._license

    @property
    def urls(self) -> List[EbrainsDatasetUrl]:
        return [{"url": self._url}]

    @property
    def description(self) -> str:
        return self._description

    def __hash__(self):
        return hash(self.id)

    def __eq__(self, o: object) -> bool:
        return hasattr(o, "id") and self.id == o.id



# Copyright 2018-2021
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .. import anchor as _anchor
from . import tabular

import pandas as pd
from textwrap import wrap
from typing import List
try:
    from typing import TypedDict
except ImportError:
    from typing_extensions import TypedDict


class GeneExpressions(
    tabular.Tabular,
    category='molecular'
):
    """
    A set gene expressions for different candidate genes
    measured inside a brain structure.
    """

    DESCRIPTION = """
    Gene expressions extracted from microarray data in the Allen Atlas.
    """

    ALLEN_ATLAS_NOTIFICATION = """
    For retrieving microarray data, siibra connects to the web API of
    the Allen Brain Atlas (© 2015 Allen Institute for Brain Science),
    available from https://brain-map.org/api/index.html. Any use of the
    microarray data needs to be in accordance with their terms of use,
    as specified at https://alleninstitute.org/legal/terms-use/.
    """

    class _DonorDict(TypedDict):
        id: int
        name: str
        race: str
        age: int
        gender: str

    class _SampleStructure(TypedDict):
        id: int
        name: str
        abbreviation: str
        color: str

    def __init__(
        self,
        levels: List[float],
        z_scores: List[float],
        genes: List[str],
        additional_columns: dict,
        anchor: _anchor.AnatomicalAnchor,
        datasets: List = []
    ):
        """
        Construct gene expression table.

        Parameters
        ----------
        levels : list of float
            Expression levels measured
        z_scores : list of float
            corresponding z scores measured
        genes : list of str
            Name of the gene corresponding to each measurement
        additional_columns : dict of list
            columns with additional data to be added to the tabular feature.
            Keys give column names, values are lists with the column data.
            Each list given needs to have the same length as expression_levels
        anchor: AnatomicalAnchor
        datasets : list
            list of datasets corresponding to this feature
        """
        assert len(z_scores) == len(levels)
        assert len(genes) == len(levels)
        if additional_columns is not None:
            assert all(len(lst) == len(levels) for lst in additional_columns.values())
        else:
            additional_columns = {}

        _data_cahced = pd.DataFrame(
            dict(
                **{'level': levels, 'zscore': z_scores, 'gene': genes},
                **additional_columns
            )
        )
        # data.index.name = 'probe_id'
        tabular.Tabular.__init__(
            self,
            description=(
                (self.DESCRIPTION + self.ALLEN_ATLAS_NOTIFICATION)
                .replace('\n', ' ')
                .replace('\t', '')
                .strip()
            ),
            modality="Gene expression",
            anchor=anchor,
            data=_data_cahced,
            datasets=datasets
        )
        self.unit = "expression level"

    def plot(self, *args, backend="matplotlib", **kwargs):
        """
        Create a box plot per gene.

        Parameters
        ----------
        backend: str
            "matplotlib", "plotly", or others supported by pandas DataFrame
            plotting backend.
        **kwargs
            Keyword arguments are passed on to the plot command.
        """
        wrapwidth = kwargs.pop("textwrap") if "textwrap" in kwargs else 40
        kwargs["title"] = kwargs.pop("title", None) \
            or "\n".join(wrap(f"{self.modality} measured in {self.anchor._regionspec}", wrapwidth))
        kwargs["kind"] = "box"
        if backend == "matplotlib":
            for arg in ['yerr', 'y', 'ylabel', 'xlabel', 'width']:
                assert arg not in kwargs
            passed_kwarg = {
                "grid": True, "legend": False, 'by': "gene", 'column': ['level'], 'showfliers': False, 'ax': None, **kwargs
            }
            return self.data.plot(*args, **passed_kwarg, backend=backend)
        elif backend == "plotly":
            kwargs["title"] = kwargs["title"].replace('\n', "<br>")
            return self.data.plot(y='level', x='gene', backend=backend, **kwargs)
        else:
            return self.data.plot(*args, backend=backend, **kwargs)



# Copyright 2018-2021
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..commons import logger, __version__, SIIBRA_USE_CONFIGURATION, siibra_tqdm
from ..retrieval.repositories import GitlabConnector, RepositoryConnector
from ..retrieval.exceptions import NoSiibraConfigMirrorsAvailableException
from ..retrieval.requests import SiibraHttpRequestError

from typing import Union
from collections import defaultdict
from requests.exceptions import ConnectionError
from os import path


class Configuration:
    """
    Provides access to siibra configurations stored in different places,
    and on request builds preconfigured objects from selected subfolders
    of the the configuration.

    Configuration repositories and detected preconfiguration spec files
    are stored globally as class variables and shared by Configuration instances.
    """

    CONFIG_REPOS = [
        ("https://jugit.fz-juelich.de", 3484),
        ("https://gitlab.ebrains.eu", 93),
    ]

    CONFIGURATIONS = [
        GitlabConnector(server, project, "siibra-{}".format(__version__), skip_branchtest=True)
        for server, project in CONFIG_REPOS
    ]

    CONFIGURATION_EXTENSIONS = []

    _cleanup_funcs = []

    @staticmethod
    def get_folders(connector: RepositoryConnector):
        return {
            path.dirname(f)
            for f in connector.search_files(suffix='.json', recursive=True)
        }

    def __init__(self):

        # lists of loaders for json specification files
        # found in the siibra configuration, stored per
        # preconfigured class name. These files can
        # loaded and fed to the Factory.from_json
        # to produce the corresponding object.
        self.spec_loaders = defaultdict(list)

        # retrieve json spec loaders from the default configuration
        for connector in self.CONFIGURATIONS:
            try:
                for folder in self.get_folders(connector):
                    loaders = connector.get_loaders(folder, suffix='.json')
                    if len(loaders) > 0:
                        self.spec_loaders[folder] = loaders
                break
            except (ConnectionError, SiibraHttpRequestError):
                logger.error(f"Cannot load configuration from {str(connector)}")
                *_, last = self.CONFIGURATIONS
                if connector is last:
                    raise NoSiibraConfigMirrorsAvailableException(
                        "Tried all mirrors, none available."
                    )
        else:
            raise RuntimeError("Cannot pull any default siibra configuration.")

        # add additional spec loaders from extension configurations
        for connector in self.CONFIGURATION_EXTENSIONS:
            try:
                for folder in self.get_folders(connector):
                    self.spec_loaders[folder].extend(
                        connector.get_loaders(folder, suffix='json')
                    )
                break
            except ConnectionError:
                logger.error(f"Cannot connect to configuration extension {str(connector)}")
                continue

        logger.debug(
            "Preconfigurations: "
            + " | ".join(f"{folder}: {len(L)}" for folder, L in self.spec_loaders.items())
        )

    @property
    def folders(self):
        return list(self.spec_loaders.keys())

    @classmethod
    def use_configuration(cls, conn: Union[str, RepositoryConnector]):
        if isinstance(conn, str):
            conn = RepositoryConnector._from_url(conn)
        if not isinstance(conn, RepositoryConnector):
            raise RuntimeError("Configuration needs to be an instance of RepositoryConnector or a valid str")
        logger.info(f"Using custom configuration from {str(conn)}")
        cls.CONFIGURATIONS = [conn]
        # call registered cleanup functions
        for func in cls._cleanup_funcs:
            func()

    @classmethod
    def extend_configuration(cls, conn: Union[str, RepositoryConnector]):
        if isinstance(conn, str):
            conn = RepositoryConnector._from_url(conn)
        if not isinstance(conn, RepositoryConnector):
            raise RuntimeError("conn needs to be an instance of RepositoryConnector or a valid str")
        if conn in cls.CONFIGURATION_EXTENSIONS:
            logger.warning(f"The configuration {str(conn)} is already registered.")
        else:
            logger.info(f"Extending configuration with {str(conn)}")
            cls.CONFIGURATION_EXTENSIONS.append(conn)
            # call registered cleanup functions
            for func in cls._cleanup_funcs:
                func()

    @classmethod
    def register_cleanup(cls, func):
        """
        Register an arbitrary function that should be executed when the
        configuration is changed, e.g. with use_configuration().
        """
        cls._cleanup_funcs.append(func)

    def build_objects(self, folder: str, **kwargs):
        """
        Build the preconfigured objects of the specified class, if any.
        """
        result = []

        if folder not in self.folders:
            logger.warning(f"No configuration found for building from configuration folder {folder}.")
            return result

        from .factory import Factory
        specloaders = self.spec_loaders.get(folder, [])
        if len(specloaders) == 0:  # no loaders found in this configuration folder!
            return result

        obj0 = Factory.from_json(
            dict(
                specloaders[0][1].data,
                **{'filename': specloaders[0][0]}
            )
        )

        for fname, loader in siibra_tqdm(
            specloaders,
            total=len(specloaders),
            desc=f"Loading preconfigured {obj0.__class__.__name__} instances"
        ):
            # filename is added to allow Factory creating reasonable default object identifiers\
            obj = Factory.from_json(dict(loader.data, **{'filename': fname}))
            result.append(obj)

        return result

    def __getitem__(self, folder: str):
        return self.build_objects(folder)


if SIIBRA_USE_CONFIGURATION:
    logger.warning(f"config.SIIBRA_USE_CONFIGURATION defined, use configuration at {SIIBRA_USE_CONFIGURATION}")
    Configuration.use_configuration(SIIBRA_USE_CONFIGURATION)



# Copyright 2018-2021
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import regional_connectivity
from hashlib import md5


class FunctionalConnectivity(
    regional_connectivity.RegionalConnectivity,
    configuration_folder="features/connectivity/regional/functional",
    category="connectivity"
):
    """Functional connectivity matrix grouped by a parcellation."""

    def __init__(self, paradigm: str, **kwargs):
        regional_connectivity.RegionalConnectivity.__init__(self, **kwargs)
        self.paradigm = paradigm

        # paradign is used to distinguish functional connectivity features from each other.
        assert self.paradigm, "Functional connectivity must have paradigm defined!"

    @property
    def id(self):
        return super().id + "--" + md5(self.paradigm.encode("utf-8")).hexdigest()



# Copyright 2018-2021
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A particular brain reference space."""


from .concept import AtlasConcept

from ..locations import Point, BoundingBox
from ..commons import logger, Species

from typing import List, TYPE_CHECKING, Union

if TYPE_CHECKING:
    from ..volumes import volume


class Space(AtlasConcept, configuration_folder="spaces"):

    def __init__(
        self,
        identifier: str,
        name: str,
        species: Union[str, Species],
        volumes: List['volume.Volume'] = [],
        shortname: str = "",
        description: str = "",
        modality: str = "",
        publications: list = [],
        datasets: list = [],
        prerelease: bool = False,
    ):
        """
        Constructs a new parcellation object.

        Parameters
        ----------
            identifier : str
                Unique identifier of the space
            name : str
                Human-readable name of the space
            species: str or Species
                Specification of the species
            volumes: list[Volume]
                list of template volumes
            shortname: str, optional
                Shortform of human-readable name
            description: str, optional
                Textual description of the parcellation
            modality  :  str or None
                Specification of the modality representing this reference space
            publications: list
                List of associated publications, each a dictionary with "doi" and/or "citation" fields
            ebrains_ids : dict
                Identifiers of EBRAINS entities corresponding to this Parcellation.
                Key: EBRAINS KG schema, value: EBRAINS KG @id
        """

        AtlasConcept.__init__(
            self,
            identifier=identifier,
            name=name,
            species=species,
            shortname=shortname,
            description=description,
            modality=modality,
            publications=publications,
            datasets=datasets,
            prerelease=prerelease,
        )
        self.volumes = volumes
        for v in self.volumes:
            v.space_info = {"@id": self.id}

    def get_template(self, variant: str = None):
        """
        Get the volumetric reference template for this space.

        Parameters
        ----------
            variant: str, optional
                Some templates are provided in different variants, e.g.
                freesurfer is available as either white matter, pial or
                inflated surface for left and right hemispheres (6 variants).
                This field could be used to request a specific variant.
                Per default, the first found variant is returned.

        Returns
        -------
            Volume
                representing the reference template, or None if not available.
        """
        tests = []
        if variant is not None:
            tests.append(lambda v: hasattr(v, 'variant') and variant.lower() in v.variant.lower())
        candidates = [v for v in self.volumes if all(t(v) for t in tests)]

        if len(candidates) == 0:
            msg = f"Volume variant {variant} not available for '{self.name}'. " \
                if variant else f"No volumes available for '{self.name}'. "
            raise RuntimeError(msg)

        if len(candidates) > 1:
            logger.info(
                f"Multiple template variants available for '{self.name}': "
                f"{', '.join(c.variant for c in candidates)}. "
                f"'{candidates[0].variant}' is chosen, but you might specify another with the 'variant' parameter."
            )

        return candidates[0]

    @property
    def provides_mesh(self):
        return any(v.provides_mesh for v in self.volumes)

    @property
    def provides_image(self):
        return any(v.provides_image for v in self.volumes)

    def __getitem__(self, slices):
        """
        Get a volume of interest specification from this space.

        Parameters
        ----------
            slices: triple of slice
                Defines the x, y and z range
        """
        if len(slices) != 3:
            raise TypeError(
                "Slice access to spaces needs to define x,y and z ranges (e.g. Space[10:30,0:10,200:300])"
            )
        point1 = [0 if s.start is None else s.start for s in slices]
        point2 = [s.stop for s in slices]
        if None in point2:
            # fill upper bounds with maximum physical coordinates
            T = self.get_template()
            shape = Point(T.get_shape(-1), None).transform(T.build_affine(-1))
            point2 = [shape[i] if v is None else v for i, v in enumerate(point2)]
        return self.get_bounding_box(point1, point2)

    def get_bounding_box(self, point1, point2):
        """
        Get a volume of interest specification from this space.

        Parameters
        ----------
            point1: 3D tuple defined in physical coordinates of this reference space
            point2: 3D tuple defined in physical coordinates of this reference space
        Returns
        -------
            BoundingBox
        """
        return BoundingBox(point1, point2, self)



# Copyright 2018-2021
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Base type of features in tabular formats."""

from zipfile import ZipFile
from .. import feature

from .. import anchor as _anchor

from ... import commons

import pandas as pd
from textwrap import wrap


class Tabular(feature.Feature):
    """
    Represents a table of different measures anchored to a brain location.

    Columns represent different types of values, while rows represent different
    samples. The number of columns might thus be interpreted as the feature
    dimension.

    As an example, receptor fingerprints use rows to represent different
    neurotransmitter receptors, and separate columns for the mean and standard
    deviations measure across multiple tissue samples.
    """

    def __init__(
        self,
        description: str,
        modality: str,
        anchor: _anchor.AnatomicalAnchor,
        data: pd.DataFrame,  # sample x feature dimension
        datasets: list = [],
        prerelease: bool = False,
        id: str = None,
    ):
        feature.Feature.__init__(
            self,
            modality=modality,
            description=description,
            anchor=anchor,
            datasets=datasets,
            prerelease=prerelease,
            id=id,
        )
        self._data_cached = data

    @property
    def data(self):
        return self._data_cached.copy()

    def _export(self, fh: ZipFile):
        super()._export(fh)
        fh.writestr("tabular.csv", self.data.to_csv())

    def plot(self, *args, backend="matplotlib", **kwargs):
        """
        Create a bar plot of a columns of the data.
        Parameters
        ----------
        backend: str
            "matplotlib", "plotly", or others supported by pandas DataFrame
            plotting backend.
        **kwargs
            takes Matplotlib.pyplot keyword arguments
        """
        wrapwidth = kwargs.pop("textwrap") if "textwrap" in kwargs else 40
        kwargs["title"] = kwargs.get(
            "title",
            "\n".join(wrap(
                f"{self.modality} in {', '.join({_.name for _ in self.anchor.regions})}",
                wrapwidth
            ))
        )
        kwargs["kind"] = kwargs.get("kind", "bar")
        kwargs["y"] = kwargs.get("y", self.data.columns[0])
        if backend == "matplotlib":
            try:
                import matplotlib.pyplot as plt
            except ImportError:
                commons.logger.error("matplotlib not available. Plotting of fingerprints disabled.")
                return None
            # default kwargs
            if kwargs.get("error_y") is None:
                kwargs["yerr"] = kwargs.get("yerr", 'std' if 'std' in self.data.columns else None)
                yerr_label = f" \u00b1 {kwargs.get('yerr')}" if kwargs.get('yerr') else ''
            kwargs["width"] = kwargs.get("width", 0.95)
            kwargs["ylabel"] = kwargs.get(
                "ylabel",
                f"{kwargs['y']}{yerr_label} {self.unit if hasattr(self, 'unit') else ''}"
            )
            kwargs["grid"] = kwargs.get("grid", True)
            kwargs["legend"] = kwargs.get("legend", False)
            ax = self.data.plot(*args, backend=backend, **kwargs)
            ax.set_title(ax.get_title(), fontsize="medium")
            ax.set_xticklabels(ax.get_xticklabels(), rotation=60, ha="right")
            plt.tight_layout()
            return ax
        elif backend == "plotly":
            kwargs["title"] = kwargs["title"].replace("\n", "<br>")
            kwargs["error_y"] = kwargs.get("error_y", 'std' if 'std' in self.data.columns else None)
            error_y_label = f" &plusmn; {kwargs.get('error_y')}" if kwargs.get('error_y') else ''
            kwargs["labels"] = {
                "index": kwargs.pop("xlabel", None) or kwargs.pop("index", ""),
                "value": kwargs.pop("ylabel", None) or kwargs.pop(
                    "value",
                    f"{kwargs.get('y')}{error_y_label} {self.unit if hasattr(self, 'unit') else ''}"
                )
            }
            fig = self.data.plot(*args, backend=backend, **kwargs)
            fig.update_layout(
                yaxis_title=kwargs["labels"]['value'],
                title=dict(
                    automargin=True, yref="container", xref="container",
                    pad=dict(t=40), xanchor="left", yanchor="top"
                )
            )
            return fig
        else:
            return self.data.plot(*args, backend=backend, **kwargs)



# Copyright 2018-2021
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Handles reading and preparing nifti files."""

from . import volume

from ..commons import logger
from ..retrieval import requests
from ..locations import pointset, boundingbox as _boundingbox

from typing import Union, Dict
import nibabel as nib
import os
import numpy as np


class NiftiProvider(volume.VolumeProvider, srctype="nii"):

    def __init__(self, src: Union[str, Dict[str, str], nib.Nifti1Image]):
        """
        Construct a new NIfTI volume source, from url, local file, or Nift1Image object.
        """
        volume.VolumeProvider.__init__(self)

        self._init_url: Union[str, Dict[str, str]] = None

        def loader(url):
            if os.path.isfile(url):
                return lambda fn=url: nib.load(fn)
            else:
                req = requests.HttpRequest(url)
                return lambda req=req: req.data

        if isinstance(src, nib.Nifti1Image):
            self._img_loaders = {None: lambda img=src: img}
        elif isinstance(src, str):  # one single image to load
            self._img_loaders = {None: loader(src)}
        elif isinstance(src, dict):  # assuming multiple for fragment images
            self._img_loaders = {lbl: loader(url) for lbl, url in src.items()}
        else:
            raise ValueError(f"Invalid source specification for {self.__class__}: {src}")
        if not isinstance(src, nib.Nifti1Image):
            self._init_url = src

    @property
    def _url(self) -> Union[str, Dict[str, str]]:
        return self._init_url

    @property
    def fragments(self):
        return [k for k in self._img_loaders if k is not None]

    @property
    def boundingbox(self):
        """
        Return the bounding box in physical coordinates
        of the union of fragments in this nifti volume.
        """
        bbox = None
        for loader in self._img_loaders.values():
            img = loader()
            if len(img.shape) > 3:
                logger.warning(
                    f"N-D NIfTI volume has shape {img.shape}, but "
                    f"bounding box considers only {img.shape[:3]}"
                )
            shape = img.shape[:3]
            next_bbox = _boundingbox.BoundingBox((0, 0, 0), shape, space=None) \
                .transform(img.affine)
            bbox = next_bbox if bbox is None else bbox.union(next_bbox)
        return bbox

    def _merge_fragments(self) -> nib.Nifti1Image:
        # TODO this only performs nearest neighbor interpolation, optimized for float types.
        bbox = self.boundingbox
        num_conflicts = 0
        result = None

        for loader in self._img_loaders.values():
            img = loader()
            if result is None:
                # build the empty result image with its own affine and voxel space
                s0 = np.identity(4)
                s0[:3, -1] = list(bbox.minpoint.transform(np.linalg.inv(img.affine)))
                result_affine = np.dot(img.affine, s0)  # adjust global bounding box offset to get global affine
                voxdims = np.asanyarray(bbox.transform(result_affine).shape, dtype="int")
                result_arr = np.zeros(voxdims, dtype=img.dataobj.dtype)
                result = nib.Nifti1Image(dataobj=result_arr, affine=result_affine)

            arr = np.asanyarray(img.dataobj)
            Xs, Ys, Zs = np.where(arr != 0)
            Xt, Yt, Zt, _ = np.split(
                (np.dot(
                    np.linalg.inv(result_affine),
                    np.dot(img.affine, np.c_[Xs, Ys, Zs, Zs * 0 + 1].T)
                ) + .5).astype('int'),
                4, axis=0
            )
            num_conflicts += np.count_nonzero(result_arr[Xt, Yt, Zt])
            result_arr[Xt, Yt, Zt] = arr[Xs, Ys, Zs]

        if num_conflicts > 0:
            num_voxels = np.count_nonzero(result_arr)
            logger.warning(f"Merging fragments required to overwrite {num_conflicts} conflicting voxels ({num_conflicts/num_voxels*100.:2.1f}%).")

        return result

    def fetch(
        self,
        fragment: str = None,
        voi: _boundingbox.BoundingBox = None,
        label: int = None
    ):
        """
        Loads and returns a Nifti1Image object

        Parameters
        ----------
        fragment: str
            Optional name of a fragment volume to fetch, if any.
            For example, some volumes are split into left and right hemisphere fragments.
            see :func:`~siibra.volumes.Volume.fragments`
        voi : BoundingBox
            optional specification of a volume of interest to fetch.
        label: int, default: None
            Optional: a label index can be provided. Then the mask of the
            3D volume will be returned, where voxels matching this label
            are marked as "1".
        """

        result = None
        if len(self._img_loaders) > 1:
            if fragment is None:
                logger.info(
                    f"Merging fragments [{', '.join(self._img_loaders.keys())}]. "
                    f"You can select one using {self.__class__.__name__}.fetch(fragment=<name>)."
                )
                result = self._merge_fragments()
            else:
                matched_names = [n for n in self._img_loaders if fragment.lower() in n.lower()]
                if len(matched_names) != 1:
                    raise ValueError(
                        f"Requested fragment '{fragment}' could not be matched uniquely "
                        f"to [{', '.join(self._img_loaders)}]"
                    )
                else:
                    result = self._img_loaders[matched_names[0]]()
        else:
            assert len(self._img_loaders) > 0
            fragment_name, loader = next(iter(self._img_loaders.items()))
            if all(s is not None for s in [fragment, fragment_name]):
                assert fragment.lower() in fragment_name.lower()
            result = loader()

        if voi is not None:
            bb_vox = voi.transform_bbox(np.linalg.inv(result.affine))
            (x0, y0, z0), (x1, y1, z1) = bb_vox.minpoint, bb_vox.maxpoint
            shift = np.identity(4)
            shift[:3, -1] = bb_vox.minpoint
            result = nib.Nifti1Image(
                dataobj=result.dataobj[x0:x1, y0:y1, z0:z1],
                affine=np.dot(result.affine, shift),
            )

        if label is not None:
            result = nib.Nifti1Image(
                (result.get_fdata() == label).astype('uint8'),
                result.affine
            )

        return result

    def get_shape(self, resolution_mm=None):
        if resolution_mm is not None:
            raise NotImplementedError(
                "NiftiVolume does not support to specify different image resolutions"
            )
        try:
            return self.image.shape
        except AttributeError as e:
            logger.error(
                f"Invalid object type {type(self.image)} of image for {self} {self.name}"
            )
            raise (e)

    def is_float(self):
        return self.image.dataobj.dtype.kind == "f"

    def find_peaks(self, min_distance_mm=5):
        """
        Find peaks in the image data.

        Parameters
        ----------
        min_distance_mm : float
            Minimum distance between peaks in mm

        Returns:
        --------
        PointSet
        """

        from skimage.feature.peak import peak_local_max
        from ..commons import affine_scaling

        img = self.fetch()
        dist = int(min_distance_mm / affine_scaling(img.affine) + 0.5)
        voxels = peak_local_max(
            img.get_fdata(),
            exclude_border=False,
            min_distance=dist,
        )
        return (
            pointset.PointSet(
                [np.dot(img.affine, [x, y, z, 1])[:3] for x, y, z in voxels],
                space=self.space,
            ),
            img,
        )


class ZipContainedNiftiProvider(NiftiProvider, srctype="zip/nii"):

    def __init__(self, src: str):
        """
        Construct a new NIfTI volume source, from url, local file, or Nift1Image object.
        """
        volume.VolumeProvider.__init__(self)
        zipurl, zipped_file = src.split(" ")
        req = requests.ZipfileRequest(zipurl, zipped_file)
        self._img_loaders = {None: lambda req=req: req.data}

        # required for self._url property
        self._init_url = src
